#include "LPC11xx.h"

#include <stdbool.h>

#include "I2CEEprom.h"

#define I2C_CON_I2EN       (1 << 6)  /* I2C Control Set Register */
#define I2C_CON_AA         (1 << 2)
#define I2C_CON_SI         (1 << 3)
#define I2C_CON_STO        (1 << 4)
#define I2C_CON_STA        (1 << 5)

#define DEFAULT_ADDRESS 0xa0
#define BAUD_100k (100 * 1000)
#define BAUD_400k (400 * 1000)

#define pDevice LPC_I2C

#define IRQ_PRIORITY_I2C 8


static void SetBaudrate(uint32_t baud)
{
	uint32_t x = SystemCoreClock;
	pDevice->SCLH = x/2;	// must be >= 4
	pDevice->SCLL = x/2;	// must be >= 4
}

static inline void Enable()
{
	pDevice->CONCLR = I2C_CON_AA|I2C_CON_SI|I2C_CON_STA;
	pDevice->CONSET = I2C_CON_I2EN;
}

static inline void Disable()
{
	pDevice->CONCLR = I2C_CON_AA|I2C_CON_SI|I2C_CON_STA|I2C_CON_I2EN;
}

void I2CEEpromInit()
{
	LPC_SYSCON->SYSAHBCLKCTRL |= (1U << 5);			// Enables clock for I2C.
	LPC_SYSCON->PRESETCTRL |= (1U << 1);			// Set Bit I2C_RST_N; enables I2C Clock

	LPC_IOCON->PIO0_4 = (1 << 0);					// Selects I2C function SCL (open-drain pin).
	LPC_IOCON->PIO0_5 = (1 << 0);					// Selects I2C function SDA (open-drain pin).

	SetBaudrate(BAUD_400k);

	NVIC_SetPriority(I2C_IRQn, IRQ_PRIORITY_I2C);
	NVIC_EnableIRQ(I2C_IRQn);
	Enable();
}

static bool mMode2ByteAddress = false;
static uint8_t mDeviceAddress = DEFAULT_ADDRESS;

static volatile uint8_t		mTransferMode;
static volatile int			mAddressIdx;
static volatile uint8_t		mControlByte;
static volatile uint8_t		mAddressBytes[2];
static volatile unsigned	mDataLength;
volatile uint8_t* volatile	pTransferBuffer;
volatile int 				mResult;

enum TransferModes {
	modeWrite,
	modeRead,
	modeCompare,
};

void I2C_IRQHandler()
{
	int res = 0;
	uint32_t clr = I2C_CON_SI;			// Clear Interrupt flag always
	uint32_t set = 0;

	switch (pDevice->STAT)
	{
	// Master general
	case 0x08:							// A START condition has been transmitted
	case 0x10:							// A repeated START condition has been transmitted
	{
		uint8_t c = mControlByte;
		// set Read Mode only after the Address write has finished
		if ((modeWrite != mTransferMode) && (mAddressIdx < 0))
			c |= 1;

		pDevice->DAT = c;
		clr |= I2C_CON_STA;
		break;
	}

	// Master transmit
	case 0x20:							// SLA+W has been transmitted; NOT ACK has been received
		res = errNotAck;
		/* no break */
	case 0x30:							// Data byte in I2DAT has been transmitted, NOT ACK has been received
		set |= I2C_CON_STO;
		break;

	case 0x38:							// Arbitration lost
		break;

	case 0x18:							// SLA+W has been transmitted; ACK has been received
	case 0x28:							// Data byte in I2DAT has been transmitted, ACK has been received
	{
		// first check, if the Address has to be sent
		int i = mAddressIdx;
		if (i > 0)
		{
			i--;
			pDevice->DAT = mAddressBytes[i];
			mAddressIdx = i;
		}
		else
		{
			// than check once, if we are in read mode
			// the check is done if AddressIdx == 0
			if (i == 0)
			{
				mAddressIdx = -1;			// check only once
				if (mTransferMode != modeWrite)
				{
					// In Read mode we have to send a repeated Start Condition
					set |= I2C_CON_STA;
					break;
				}
			}

			// than check if there are Bytes to send
			unsigned l = mDataLength;
			if (l)
			{
				pDevice->DAT = *pTransferBuffer++;
				mDataLength = --l;
			}
			else
			{
				// no more Bytes; send Stop Condition
				set |= I2C_CON_STO;
			}
		}
		break;
	}

	// Master receive
	case 0x40:							// SLA+R has been transmitted; ACK has been received
	{
		unsigned l = mDataLength;
		if (l) set |= I2C_CON_AA;		// if there are more bytes to receive, send ACK
		else clr |= I2C_CON_AA;			// else NACK
		break;
	}

	case 0x50:							// Data byte has been received; ACK has been returned
	{
		unsigned l = mDataLength;
		if (l)
		{
			if (modeRead == mTransferMode)
			{
				*pTransferBuffer++ = pDevice->DAT;
			}
			else if (modeCompare == mTransferMode)
			{
				// don't write into pTransferBuffer at compare!
				if (*pTransferBuffer == (uint8_t)pDevice->DAT)
				{
					pTransferBuffer++;	// if compare is OK increment Buffer
				}
				else
				{
					l = 1;				// else stop receiving early
				}
			}
			mDataLength = --l;
		}
		if (l) set |= I2C_CON_AA;		// if there are more bytes to receive, send ACK
		else clr |= I2C_CON_AA;			// else NACK
	}
	break;

	case 0x48:							// SLA+R has been transmitted; NOT ACK has been received
		res = errNotAck;
		/* no break */
	case 0x58:							// Data byte has been received; NOT ACK has been returned
		set |= I2C_CON_STO;
		break;

	case 0:								// Bus Error
		res = errBus;
		set |= I2C_CON_STO;
		break;
	}
	// first do the set operation
	pDevice->CONSET = set;
	// finally do the reset since reset of the SI bit will start the next action
	pDevice->CONCLR = clr;

	if (set & I2C_CON_STO)
	{
		// negative = error, positive = success, 0 = work in progress
		if (res == 0) res = 1;
		mResult = res;
	}
}

static void PrepareCommunication(uint16_t eeAddr, const void* buff, unsigned len, uint8_t mode)
{
	mTransferMode = mode;
	mDataLength = len;
	pTransferBuffer = (uint8_t*)buff;

	mAddressBytes[0] = eeAddr & 0xff;
	if (mMode2ByteAddress)
	{
		mAddressBytes[1] = (eeAddr >> 8) & 0xff;
		mAddressIdx = 2;
		mControlByte = mDeviceAddress & 0xf7;
	}
	else
	{
		mAddressIdx = 1;
		// The upper 3 address bits are bits 1-3 in the control byte.
		mControlByte = (mDeviceAddress & 0xf0) | ((eeAddr >> 7) & 0x0e);
	}

	mResult = 0;
	pDevice->CONSET = I2C_CON_STA;	// this will start Communication
}

static int WaitForCommunication(const void* buff)
{
	while (1)
	{
		int res = mResult;
		if (res)
		{
			// positive: number of handled bytes; negative: error
			if (res > 0)
			{
				// since the pTransferBuffer is incremented at each byte action
				// it's value minus the original value is the number of bytes.
				res = pTransferBuffer - (uint8_t*)buff;
			}
			return res;
		}

		// wait
		__WFI();
	}
}


int I2CEEPromRead(uint16_t eeAddr, void* buff, uint16_t len)
{
	PrepareCommunication(eeAddr, buff, len, modeRead);
	return WaitForCommunication(buff);
}

int I2CEEPromWrite(uint16_t eeAddr, const void* buff, uint16_t len)
{
	PrepareCommunication(eeAddr, buff, len, modeWrite);
	return WaitForCommunication(buff);
}

int I2CEEPromCompare(uint16_t eeAddr, const void* buff, uint16_t len)
{
	PrepareCommunication(eeAddr, buff, len, modeCompare);
	// return number of equal Bytes
	// ISR stops incrementing buffer if compare is not OK,
	// so the BufferPos related to the given Buffer is the number of equal bytes
	return WaitForCommunication(buff);
}
